/*
 * Copyright © 2006 Carl Worth
 *
 * This program is free software; you can redistribute it and\/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA."
 */

#include "word-game.h"

#include <string.h>
#include <ctype.h>
#include <sys/time.h>
#include <time.h>

char bag_orig[101] =
"AAAAAAAAABBCCDDDDEEEEEEEEEEEEFFGGGHHIIIIIIIIIJKLLLL"
"MMNNNNNNOOOOOOOOPPQRRRRRRSSSSTTTTTTUUUUVVWWXYYZ??";

static int
rand_within (int num_values)
{
    return (int) ((double) num_values * (rand() / (RAND_MAX + 1.0)));
}

static void
shuffle (char *array, int length)
{
    int i, r, tmp;

    for (i = 0; i < length; i++)
    {
	r = i + rand_within (length - i);
	tmp = array[i];
	array[i] = array[r];
	array[r] = tmp;
    }
}

static const char *subanagram_anagram;
static char subanagram_enumerate_word[WORD_GAME_MAX_WORD_LENGTH];
static dict_t *subanagram_enumerate_answers;

static void
subanagram_enumerate (dict_cursor_t	 cursor,
		      uint8_t		 seen)
{
    char c;
    unsigned int i, last;
    uint8_t next_seen;

    if (cursor == DICT_CURSOR_NIL)
	return;

    if (DICT_ENTRY_IS_WORD (dict_cursor_resolve (cursor)))
	dict_add_word (subanagram_enumerate_answers, subanagram_enumerate_word);

    last = strlen (subanagram_enumerate_word);

    for (i = 0; i < strlen (subanagram_anagram); i++) {
	if (seen & (1 << i))
	    continue;
	next_seen = seen | (1 << i);
	c = subanagram_anagram[i];
	if (c == '?' || c == '_') {
	    for (c = 'a'; c <= 'z'; c++) {
		subanagram_enumerate_word[last] = c;
		subanagram_enumerate (dict_cursor_next (cursor, c), next_seen);
	    }
	} else {
	    subanagram_enumerate_word[last] = c;
	    subanagram_enumerate (dict_cursor_next (cursor, c), next_seen);
	}
	subanagram_enumerate_word[last] = '\0';
    }
}

static void
subanagram_expand (const char *anagram, dict_t *dict, dict_t *answers)
{
    subanagram_anagram = anagram;
    memset (subanagram_enumerate_word, '\0', WORD_GAME_MAX_WORD_LENGTH);
    subanagram_enumerate_answers = answers;

    subanagram_enumerate (dict_root (dict), 0);
}

static int
character_compare (const void *_a,
		   const void *_b)
{
    const char *a = _a;
    const char *b = _b;

    if (*a == '?')
	return 1;
    if (*b == '?')
	return -1;
    return *a - *b;
}

int
main (void)
{
    dict_t dict, solution;
    struct timeval tv;
    char bag[100], rack[8];

    gettimeofday (&tv, NULL);
    srand (tv.tv_sec ^ tv.tv_usec);

    memcpy (bag, bag_orig, 100);
    shuffle (bag, 100);

    memcpy (rack, bag, 7);
    rack[7] = '\0';
    qsort (rack, 7, 1, character_compare);

    dict_init (&dict);
    dict_add_words_from_file (&dict, "words.txt");

    dict_init (&solution);
    subanagram_expand (rack, &dict, &solution);

    word_game_play (rack, &dict, &solution, 0);

    dict_fini (&solution);
    dict_fini (&dict);

    return 0;
}
